/* ------------------------------------------------------------------------ */
/*  STRQUOTE.C  --  Performes "quoting" of strings to make them storable    */
/*                  and transmittable in text format, such as in an .INI    */
/*                  file.                                                   */
/* ------------------------------------------------------------------------ */
/*  Quoting scheme used:                                                    */
/*  All characters are translated as-is except:                             */
/*                                                                          */
/*  ^ quote-character, can be followed by:                                  */
/*  @       -> NUL (Ctrl-@, ASCII 0)                                        */
/*  A..Z    -> SOH..SUB (Ctrl-A to Ctrl-Z, ASCII 1-26)                      */
/*  [       -> ESC (Ctrl-[, ASCII 27)                                       */
/*  \       -> FS  (Ctrl-\, ASCII 28)                                       */
/*  ]       -> GS  (Ctrl-], ASCII 29)                                       */
/*  ^       -> Caret character (^, ASCII 94)    NOTE!                       */
/*  _       -> US  (Ctrl-_, ASCII 31)                                       */
/*  a..z    -> SOH..SUB                                                     */
/*  {       -> ESC                                                          */
/*  |       -> FS                                                           */
/*  }       -> GS                                                           */
/*  ~       -> RS  (Ctrl-^, ASCII 30)                                       */
/*                                                                          */
/*  ~ quote-character, can be followed by:                                  */
/*  Hex d.  -> The ASCII code specified by two hex digits                   */
/*  ~       -> Tilde character (~, ASCII 126)                               */
/*  ^       -> Caret character (^, ASCII 94)                                */
/*  @       -> DEL (ASCII 127)                                              */
/*  _       -> Space character (ASCII 32)                                   */
/*                                                                          */
/*  En-quote mode flags in _QuoteMode:                                      */
/*  QTE_DEF     - Quote characters 0-31, 94, 126-127 and 255 only.          */
/*  QTE_HICTRL  - Quote characters 128-159 as well.                         */
/*  QTE_HIGH    - Quote all characters 0-255.                               */
/*  QTE_ESPC    - Quote spaces which are last characters (default).         */
/*  QTE_SPC     - Quote all spaces.                                         */
/*  QTE_NOSPC   - Do not quote spaces.                                      */
/*																			*/
/*	De-quote mode flags in _QuoteMode:										*/
/*	QTE_STR		- Treat output data as a string (add a zero byte, default).	*/
/*	QTE_MEM		- Treat output data as binary (add no zero byte).			*/
/* ------------------------------------------------------------------------ */

#include <stdlib.h>

#define QTE_DEF     (0x00)
#define QTE_HICTRL  (0x01)
#define QTE_HIPR    (0x02)
#define QTE_HIGH	(0x03)      /* QTE_HIGH == QTE_HICTRL|QTE_HIPR      */
#define QTE_ESPC    (0x00)
#define QTE_SPC     (0x04)
#define QTE_NOSPC   (0x08)
#define QTE_STR		(0x00)
#define QTE_MEM		(0x10)

int _QuoteMode = QTE_DEF;


/* ------------------------------------------------------------------------ */
/*  char *quotestr(char *outstr, const void *in, size_t len)           		*/
/*                                                                          */
/*  Converts a string to quoted form.  Since the string can contain zeroes, */
/*  a length parameter is included.  The global variable _QuoteMode is set  */
/*  to the desired quoting mode (it is assumed to be rarely changed).       */
/* ------------------------------------------------------------------------ */

char *quotestr(char *outstr, const void *in, size_t len)
{
#define instr	((char *)in)

    static const char hexdig[16] = "0123456789ABCDEF";
    char *out, ch;

    out = outstr;

    while ( len-- )
    {
        ch = *(instr++);

        if ( ch <= '\x1F' && ch >= '\0' )
        {
            *(out++) = '^';
            *(out++) = ( ch == '\x1E' ? '~' : ch+'@' );
        }
        else if ( ch >= '\x80' && ch <= '\xFF' )
        {
			if ( _QuoteMode & ( ch <= '\x9F' ? QTE_HICTRL : QTE_HIPR ) || ch == '\xFF' )
            {
                *(out++) = '~';
                *(out++) = hexdig[ch>>4];
                *(out++) = hexdig[ch&0x0F];
            }
            else
                *(out++) = ch;
        }
        else if ( ch == ' ' )
        {
			if ( _QuoteMode & QTE_SPC || ((_QuoteMode & QTE_NOSPC)==0 && len == 0) )
            {
                *(out++) = '~';
                *(out++) = '_';
            }
            else
                *(out++) = ch;
        }
        else if ( ch == '^' || ch == '~' )
        {
            *(out++) = ch;
            *(out++) = ch;
        }
        else if ( ch == '\x7F' )
        {
            *(out++) = '~';
            *(out++) = '@';
        }
        else
            *(out++) = ch;
    }
	*out = '\0';

    return outstr;

#undef instr
}


/* ------------------------------------------------------------------------ */
/*  size_t unquotestr(void *outstr, const char *instr)                      */
/*                                                                          */
/*  Decodes a quoted string.  Returns the length of the output string,      */
/*  which may contain nulls.  An extra null is appended to the end of the   */
/*  string but is not included in the length.                               */
/* ------------------------------------------------------------------------ */

size_t unquotestr(void *out, const char *instr)
{
#define outstr	((char *)out)

	size_t len;
	char ch, hxbuf[3];

	hxbuf[2] = '\0';

	len = 0;

    while ( (ch = *(instr++)) != '\0' )
    {
        switch(ch)
        {
            case '^':
                        ch = *(instr++);

                        if ( ch >= '@' && ch <= '~' && ch != '^' )
                        {
							*(outstr++) = ch & 0x1F;
                            len++;
                        }
                        else if ( ch == '\0' )
                        {
                            instr--;
                        }
                        else
                        {
                            *(outstr++) = ch;
                            len++;
                        }
                        break;

            case '~':
                        ch = *(instr++);
                        switch(ch)
                        {
                            case '0':
                            case '1':
                            case '2':
                            case '3':
                            case '4':
                            case '5':
                            case '6':
                            case '7':
                            case '8':
                            case '9':
                            case 'A':
                            case 'B':
                            case 'C':
                            case 'D':
                            case 'E':
                            case 'F':
                            case 'a':
                            case 'b':
                            case 'c':
                            case 'd':
                            case 'e':
                            case 'f':
                                        hxbuf[0] = ch;
                                        hxbuf[1] = *instr;

                                        *(outstr++) = (char) strtol(hxbuf,NULL,16);
                                        len++;

                                        if ( *instr )
                                            instr++;

                                        break;

                            case '@':
                                        *(outstr++) = '\x7F';
                                        len++;
                                        break;

                            case '_':
                                        *(outstr++) = ' ';
                                        len++;
                                        break;

                            case '\0':
                                        instr--;
                                        break;

                            default:
                                        *(outstr++) = ch;
                                        len++;
                                        break;
                        }
                        break;

            default:
						*(outstr++) = ch;
						len++;
                        break;
        }
	}

	if ( (_QuoteMode & QTE_MEM) == 0 )
		*outstr = '\0';

	return len;

#undef outstr
}
